/*
 * Copyright (c) 2008 by Hartmut Birr
 *
 * This program is free software; you can redistribute it and/or
 * mmodify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA. 
 *
 */

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <util/crc16.h>
#include <util/delay.h>

#include <stdio.h>
#include <string.h>

#define BUFFER_SIZE 19          // 6 channels * 3 colors + a overflow/reload entry
#define BAUD_RATE   115200

typedef struct
{
    uint8_t b;
    uint8_t c;
    uint8_t d;
}
port_t;

typedef struct
{
    uint16_t timer;
    port_t port;
} pwm_t;

static char version[] PROGMEM = "aurora 0.0.1-beta (" __DATE__ " " __TIME__ ")";

static uint8_t id;

static pwm_t buffer_int[3 * BUFFER_SIZE];

static pwm_t* buffer_0_ptr;
static pwm_t* buffer_1_ptr;
static pwm_t* buffer_2_ptr;

pwm_t* ptr_int;
pwm_t* ptr_next;

#define PULL_UP_PORT_B  0
#define PULL_UP_PORT_C  0
#define PULL_UP_PORT_D  0

static uint16_t init_gamatab[33] PROGMEM = 
{
        0,     4,    24,    66,   136,   237,   374,   550,
      768,  1031,  1342,  1703,  2116,  2585,  3111,  3697,
     4344,  5055,  5832,  6676,  7589,  8574,  9631, 10764, 
    11972, 13258, 14624, 16071, 17601, 19215, 20914, 22701, 
    24576
};

static uint16_t eep_gammatab[33] EEMEM =   
{
        0,     4,    24,    66,   136,   237,   374,   550, 
      768,  1031,  1342,  1703,  2116,  2585,  3111,  3697, 
     4344,  5055,  5832,  6676,  7589,  8574,  9631, 10764, 
    11972, 13258, 14624, 16071, 17601, 19215, 20914, 22701, 
    24576
};

static uint16_t gammatab[33];

static uint8_t init_config[6] PROGMEM = {0, 1, 2, 3, 4, 5};
static uint8_t eep_config[6] EEMEM    = {0, 1, 2, 3, 4, 5};
static uint8_t config[6];


#if 0
#define FILLER  0x55aa

uint16_t dummy[1215-32-3] PROGMEM = 
{
    FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, 
    FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, 
    FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, 
    FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, 
    FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, 
    FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, 
    FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, 
    FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, 
    FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, 
    FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, 
    FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, 
    FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, 
    FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, 
    FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, 
    FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, 
    FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, FILLER, 
};
#endif

static uint16_t init_whitetab[6][3] PROGMEM =
{
    {0xffff, 0xffff, 0xffff},
    {0xffff, 0xffff, 0xffff},
    {0xffff, 0xffff, 0xffff},
    {0xffff, 0xffff, 0xffff},
    {0xffff, 0xffff, 0xffff},
    {0xffff, 0xffff, 0xffff}
};

static uint16_t eep_whitetab[6][3] EEMEM =
{
    {0xffff, 0xffff, 0xffff},
    {0xffff, 0xffff, 0xffff},
    {0xffff, 0xffff, 0xffff},
    {0xffff, 0xffff, 0xffff},
    {0xffff, 0xffff, 0xffff},
    {0xffff, 0xffff, 0xffff}
};

static uint16_t whitetab[6][3];

static const port_t rgb_channel_mask[6][3] = 
{
    {
        {.d = 0x04, .c = 0x00, .b = 0x00},  // r0 = PD2
        {.d = 0x08, .c = 0x00, .b = 0x00},  // g0 = PD3
        {.d = 0x10, .c = 0x00, .b = 0x00}   // b0 = PD4
    },
    {
        {.d = 0x00, .c = 0x20, .b = 0x00},  // r1 = PC5
        {.d = 0x00, .c = 0x10, .b = 0x00},  // g1 = PC4
        {.d = 0x00, .c = 0x08, .b = 0x00}   // b1 = PC3
    },
    {
        {.d = 0x00, .c = 0x04, .b = 0x00},  // r2 = PC2
        {.d = 0x00, .c = 0x02, .b = 0x00},  // g2 = PC1
        {.d = 0x00, .c = 0x01, .b = 0x00}   // b2 = PC0
    },
    {
        {.d = 0x20, .c = 0x00, .b = 0x00},  // r3 = PD5
        {.d = 0x40, .c = 0x00, .b = 0x00},  // g3 = PD6
        {.d = 0x80, .c = 0x00, .b = 0x00}   // b3 = PD7
    },
    {
        {.d = 0x00, .c = 0x00, .b = 0x20},  // r4 = PB5                                                                                 
        {.d = 0x00, .c = 0x00, .b = 0x10},  // g4 = PB4
        {.d = 0x00, .c = 0x00, .b = 0x08},  // b4 = PB3
    },
    {
        {.d = 0x00, .c = 0x00, .b = 0x04},  // r5 = PB2
        {.d = 0x00, .c = 0x00, .b = 0x02},  // g5 = PB1
        {.d = 0x00, .c = 0x00, .b = 0x01}   // b5 = PB0
    }
};

static uint16_t rgb_timer_values[6][3];

#define MSG_MULTICAST   0x80
#define MSG_ANSWER      0x40
#define MSG_MASK        0x3F

#define ID_ERR_CRC      0x80
#define ID_ERR_CMD      0x40
#define ID_ERR_PARAM    0x20
#define ID_MASK         0x1F

#define MSG_START       0xAA
#define MSG_ESCAPE      0x55

static union
{
    uint8_t buffer[255+4];
    struct
    {
        uint8_t cmd;
        uint8_t id;
        uint8_t len;
        uint8_t data[255];
        uint8_t crc;
    };
}data;

static uint8_t crc;

void getMessage(void)
{
    uint16_t pos;
    uint8_t ch;
    uint8_t escape;
    uint8_t tmp_crc;

WaitForStart:

    do
    {
        while (!(UCSR0A & (1<<RXC0)));
        ch = UDR0;
    } while (ch != MSG_START);

Restart:

    tmp_crc = _crc_ibutton_update(0, MSG_START);
    pos = 0;
    escape = 0;

WaitForNext:

    while (!(UCSR0A & (1<<RXC0)));
    ch = UDR0;

    if (ch == MSG_START)
    {
        goto Restart;
    }

NextEscape:

    if (escape == 1)
    {
        if (ch == 0)
        {
            ch = MSG_ESCAPE;
            escape = 0;
        }
        else if (ch == 1)
        {
            ch = MSG_START;
            escape = 0;
        }
        else if (ch == 2)
        {
            ch = MSG_ESCAPE;
            escape = 2;
        }
        else if (ch == 3)
        {
            ch = MSG_START;
            escape = 3;
        }
        else if (ch >= 0x12 && ch <= 0x1f)
        {
            escape = ch - 1;
            if (pos > 0)
            {
                ch = data.buffer[pos - 1];
            }
            else
            {
                ch = 0xaa;
            }
        }
        else
        {
            goto WaitForStart;
        }
    }
    else if (escape == 2)
    {
        ch = MSG_START;
        escape = 0;
    }
    else if (escape == 3)
    {
        ch = MSG_ESCAPE;
        escape = 0;
    }
    else if (escape > 0x10 && escape <= 0x1e)
    {
        if (--escape <= 0x10)
        {
            escape = 0;
        }
    }
    else if (escape)
    {
        goto WaitForStart;
    }
    else if (ch == MSG_ESCAPE)
    {
        escape = 1;
        goto WaitForNext;
    }

    if (pos <= 2 || (pos < data.len + 3 && pos < 255 + 3))
    {
        tmp_crc = _crc_ibutton_update(tmp_crc, ch);
        data.buffer[pos] = ch;
        pos++;
        if (escape > 0)
        {
            goto NextEscape;
        }
        goto WaitForNext;
    }
    data.crc = ch;
    crc = tmp_crc;
}

void putMessage(void)
{
    uint16_t i;
    uint8_t ch;
    uint8_t crc;

    // insert a gap of one character 
    _delay_ms(10000.0 / BAUD_RATE);

    while (!(UCSR0A & (1<<UDRE0)));
    UDR0 = MSG_START;
    crc = _crc_ibutton_update(0, MSG_START);

    for (i = 0; i < data.len + 4; i++)
    {
        if (i == data.len + 3)
        {
            ch = crc;
        }
        else
        {
            ch = data.buffer[i];
            crc = _crc_ibutton_update(crc, ch);
        }

        if (ch == MSG_START || ch == MSG_ESCAPE)
        {
            while (!(UCSR0A & (1<<UDRE0)));
            UDR0 = MSG_ESCAPE;
            ch = ch == MSG_ESCAPE ? 0 : 1;
        }
        while (!(UCSR0A & (1<<UDRE0)));
        UDR0 = ch;
    }
}

uint16_t gammacorr(uint8_t channel, uint8_t rgb, uint8_t val)
{
    uint32_t index = val / 8;
    uint32_t offset = (val % 8);
    uint32_t a, b;
    uint32_t c;

    if (channel >= 6 || rgb >= 3)
    {
		return 0;
    }

    a = gammatab[index]; 
    b = gammatab[index + 1];
    c = a + offset * (b - a) / 8;
    c *= whitetab[channel][rgb];
    c /= 0xffff;
    c = ((c + 3) / 6) * 6;
    return c;
}


pwm_t* calcTimerValues(void)
{
    pwm_t* ptr;
    uint8_t i, k, m;
    uint8_t last;
    port_t on;
    port_t mask;
    uint8_t sreg;

    last = 0;
    on.b = on.c = on.d = 0;

	sreg = SREG;
	cli();
	ptr = ptr_int;  // get the buffer which is used currently
	SREG = sreg;

	// We are using three buffers:
	//   the 1th is currently used by the interrupt
	//   the 2th is pre initialized and is used if one periode is finished (timer == 0xffff)
	//   the 3th is for preperation of new datas
	// It is possible, that the 1th and 2th buffer is the same. In this case the
	// the timers didn't change.
	// ptr_int -> pointer to the current buffer (1th)
	// next_int -> index to the next buffer with pre initialized datas (2th)
	// next_buffer -> pointer to an empty buffer, which must be filled with new datas

	// Set ptr to the empty buffer
	if (!(ptr >= buffer_0_ptr && ptr < buffer_0_ptr + BUFFER_SIZE) &&
	    !(ptr_next >= buffer_0_ptr && ptr_next < buffer_0_ptr + BUFFER_SIZE))
	{
	    ptr = buffer_0_ptr;
	}
	else if (!(ptr >= buffer_1_ptr && ptr < buffer_1_ptr + BUFFER_SIZE) &&
	         !(ptr_next >= buffer_1_ptr && ptr_next < buffer_1_ptr + BUFFER_SIZE))
	{
	    ptr = buffer_1_ptr;
	}
	else 
	{
	    ptr = buffer_2_ptr;
	} 

    for (i = 0; i < 6; i++)
    {
        for (k = 0; k < 3; k++)
        {
            uint16_t timer = rgb_timer_values[i][k];
                
            if (timer > 0)
            {
                mask = rgb_channel_mask[i][k];

                on.b |= mask.b;
                on.c |= mask.c;
                on.d |= mask.d;

                if (timer < 0x5fff)
                {
                    uint8_t found = 0;

                    // find the position, where the timer should be inserted
                    for (m = 0; m < last; m++)
                    {
                        if (timer == ptr[m].timer)
                        {
			    	        // the timer does exist already, update the port mask only
			        		ptr[m].port.b |= mask.b;
			        		ptr[m].port.c |= mask.c;
			        		ptr[m].port.d |= mask.d;
							found = 1; 
							break;
                        }
			    		else if (timer < ptr[m].timer)
			    		{
			    		    // the timer doesn't exist, 
							// move all larger timers backwarts and insert the new timer
							memmove(&ptr[m+1], &ptr[m], (last-m) * sizeof(pwm_t));
			    		    ptr[m].timer = timer;
			        		ptr[m].port.b = mask.b;
			        		ptr[m].port.c = mask.c;
			        		ptr[m].port.d = mask.d;
							last++;
							found = 1;
							break;
                        }
                    }
					if (!found)
					{
			    	    // the new timer is the largest one, append the timer at the end
			    		ptr[last].timer = timer;
			        	ptr[last].port.b = mask.b;
			        	ptr[last].port.c = mask.c;
			        	ptr[last].port.d = mask.d;
			    		last++;
                    }
                }
            }
        }
    }

    // appand the timer which does switch off all LED's
    ptr[last].timer = 0xffff;
	ptr[last].port.b = PULL_UP_PORT_B;
	ptr[last].port.c = PULL_UP_PORT_C;
	ptr[last].port.d = PULL_UP_PORT_D;

    // merge the bit masks of the timers
	for (i = 0; i < last; i++)
	{
	    mask.b = ptr[i].port.b;
	    mask.c = ptr[i].port.c;
	    mask.d = ptr[i].port.d;
	    		
        ptr[i].port.b = on.b | PULL_UP_PORT_B;
        ptr[i].port.c = on.c | PULL_UP_PORT_C;
        ptr[i].port.d = on.d | PULL_UP_PORT_D;
		    		
        on.b &= ~mask.b;
        on.c &= ~mask.c;
        on.d &= ~mask.d;
	}

	// swab the timer bytes
    for (i = 0; i <= last; i++) 
	{
	    ptr[i].timer = (ptr[i].timer >> 8) | (ptr[i].timer << 8);
	}
    return ptr;
}

pwm_t* getRgbValues(uint8_t addr)
{
    uint8_t rgb_group;
    uint8_t i, j;
    uint8_t modified = 0;
    pwm_t* ptr = NULL;

    if (addr)
    {
        // logical address
        for (i = 0; i < data.len / 4; i++)
        {
            for (rgb_group = 0; rgb_group < 6; rgb_group++)
            {
                if (config[rgb_group] == data.data[4 * i])
                {
                    for (j = 0; j < 3; j++)
                    {
                        rgb_timer_values[rgb_group][j] = gammacorr(rgb_group, j, data.data[4 * i + 1 + j]);   
                    }   
                    modified = 1;
                }
            }
        }
    }
    else
    {
        // physical address
        for (i = 0; i < data.len / 4; i++)
        {
            if ((data.data[4 * i] / 8) == id)
            {
                rgb_group = data.data[4 * i] % 8;
                if (rgb_group < 6)
                {
                    for (j = 0; j < 3; j++)
                    {
                        rgb_timer_values[rgb_group][j] = gammacorr(rgb_group, j, data.data[4 * i + 1 + j]);   
                    }   
                    modified = 1;
                }
            }
        }
    }

    if (modified)
    {
        ptr = calcTimerValues();
    }
    return ptr;
}

void saveEeprom(uint8_t* dst, uint8_t* src, uint8_t size)
{
    while(size--)
    {
        if (*src != eeprom_read_byte(dst))
        {
            eeprom_write_byte(dst, *src);
        }
        dst++;
        src++;
    }
}
            

void __attribute__((noreturn)) main(void)
{
    pwm_t* ptr = NULL;
    uint8_t multicastSequence = 0;
    uint8_t multicastCmd = 0;
    uint8_t multicastResult = 0;

    sei();

    buffer_0_ptr = buffer_int;
    buffer_1_ptr = buffer_int + BUFFER_SIZE;
    buffer_2_ptr = buffer_int + 2 * BUFFER_SIZE;

    buffer_0_ptr[0].timer = 0xffff;
    buffer_0_ptr[0].port.b = PULL_UP_PORT_B;
    buffer_0_ptr[0].port.c = PULL_UP_PORT_C;
    buffer_0_ptr[0].port.d = PULL_UP_PORT_D;

    ptr_int = ptr_next = buffer_0_ptr;

    PORTB = 0;
    PORTC = 0;
    PORTD = 0;

    DDRB = (1<<DDB5)|(1<<DDB4)|(1<<DDB3)|(1<<DDB2)|(1<<DDB1)|(1<<DDB0);
    DDRC = (1<<DDC5)|(1<<DDC4)|(1<<DDC3)|(1<<DDC2)|(1<<DDC1)|(1<<DDC0);
    DDRD = (1<<DDD7)|(1<<DDD6)|(1<<DDD5)|(1<<DDD4)|(1<<DDD3)|(1<<DDD2);

    // PWM frequency = F_CPU / 48 / 4096
    // F_CPU = 14.7456MHz ->  75Hz
    // F_CPU = 16MHz      ->  81Hz
    // F_CPU = 18.432MHz  ->  94Hz
    // F_CPU = 20MHz      -> 102Hz
    OCR1A = 0xffff;
	OCR1B = 0x5fff;
	ICR1 = 0x5fff;
    TCCR1A = 0;
    TCCR1B = (1<<WGM13)|(1<<WGM12)|(0<<CS12)|(1<<CS11)|(0<<CS10);
    TIMSK1 |= (1<<OCIE1B)|(1<<OCIE1A);

    UBRR0L = (F_CPU / BAUD_RATE / 16 - 1) % 256;
    UBRR0H = (F_CPU / BAUD_RATE / 16 - 1) / 256;
    UCSR0A = 0;
    UCSR0C = (1<<UCSZ01)|(1<<UCSZ00);
    UCSR0B = (1<<RXEN0)|(1<<TXEN0);

    // power down unused peripherals
    PRR = (1<<PRTWI)|(1<<PRTIM2)|(1<<PRTIM0)|(1<<PRSPI)|(1<<PRADC);

    eeprom_read_block(whitetab, eep_whitetab, sizeof(whitetab));
    eeprom_read_block(gammatab, eep_gammatab, sizeof(gammatab));
    eeprom_read_block(config, eep_config, sizeof(config));

    if (gammatab[0] != 0 || gammatab[32] != 0x6000)
    {
   		memcpy_P(whitetab, init_whitetab, sizeof(whitetab));
		memcpy_P(gammatab, init_gamatab, sizeof(gammatab));
        memcpy_P(config, init_config, sizeof(config));
		eeprom_write_block(whitetab, eep_whitetab, sizeof(whitetab));
		eeprom_write_block(gammatab, eep_gammatab, sizeof(gammatab));
        eeprom_write_block(config, eep_config, sizeof(config));
    }

    id = pgm_read_byte(FLASHEND);

    while(1)
    {
        getMessage();

//  cmd format: <0xaa><cmd><id/sequence><len><data 0> ... <data n><crc>
//   len may be 0. In this case, there aren't send data 0..n. 
//  unicast cmd: cmd is < 0x80, <id/sequence> is the id of the slave which should response
//   0x10 -> Ready
//   0x11 -> Result of last multicast
//   0x12 -> get white calibration
//   0x13 -> set white calibration
//   0x14 -> store white calibration in eeprom
//   0x15 -> get gamma correction
//   0x16 -> get gamma correction
//   0x17 -> store gamma correction in eeprom
//   0x18 -> get versions string
//   0x19 -> get logical address configuration
//   0x1a -> set logical address configuration
//   0x1b -> store logical address configuration in eeprom
//   0x30 -> performance test, calculate 1000x the crc
//   0x31 -> performance test, calculate 5x new rgb values
//  multicast cmd: cmd is >= 0x80, <id/sequence> is the sequence no. from host
//   0x90 -> reboot
//   0x91 -> rgb information, physical addressing, update imediatly
//   0x92 -> rgb information, physical addressing, calculate values only
//   0x93 -> activate new rgb values
//   0x94 -> rgb information, logical addressing, update imediatly
//   0x95 -> rgb information, logical addressing, calculate values only
//  answer format: <0xaa><cmd><status+id/sequence><len><data 0> ... <data n><crc>  
//   len may be 0. In this case, there aren't send data 0..n. 

        if ((data.cmd & ~MSG_MASK) == MSG_MULTICAST)
        {
            // multicast
            multicastSequence = data.id & ID_MASK;
            multicastCmd = data.cmd;
            multicastResult = 0;

            if (data.crc != crc)
            {
                multicastSequence |= ID_ERR_CRC;
            }
            else
            {
                switch(data.cmd)
                {
                    case 0x90:  // restart
                        cli();
                        PORTB = PULL_UP_PORT_B;
                        PORTC = PULL_UP_PORT_C;
                        PORTD = PULL_UP_PORT_D;
                        wdt_enable(WDTO_15MS);
                        while(1);

                    case 0x91:  // rgb information, update imediatly 
                    case 0x94:
                        ptr = getRgbValues(data.cmd == 0x91 ? 0 : 1);
                        // fall throught
                    case 0x93:  // activate new values
                        if (ptr)
                        {
                            uint8_t sreg;
                            // activate the new data
                          	sreg = SREG;
				            cli();
				            ptr_next = ptr;
				            SREG = sreg;

                            ptr = NULL;
                        }
                        break;

                    case 0x92:  // rgb information, calc values only
                    case 0x95:
                        ptr = getRgbValues(data.cmd == 0x92 ? 0 : 1);
                        break;

                    case 0x96:  // set UBR
                        if (data.len < 2)
                        {
                            multicastSequence |= ID_ERR_PARAM;
                        }
                        else
                        {
                            UCSR0B = 0;
                            UBRR0L = data.data[0];
                            UBRR0H = data.data[1] & 0x0f;
                            UCSR0A = data.data[1] & 0x80 ? (1<<U2X0) : 0;
                            UCSR0C = (1<<UCSZ01)|(1<<UCSZ00);
                            UCSR0B = (1<<RXEN0)|(1<<TXEN0);
                        }
                        break;

                    default:
                        multicastSequence |= ID_ERR_CMD;
                }
            }
        }
        else if ((data.cmd & ~MSG_MASK) == 0 &&
                 data.id == id)
        {
            // unicast with our ID
            uint8_t len = data.len;
            data.len = 0;

            if (data.crc != crc)
            {
                data.id |= ID_ERR_CRC;
            }
            else
            {
                switch(data.cmd)
                {
                    default: // unknown command
                        data.id |= ID_ERR_CMD;
                        break;

                    case 0x10: // Ready
                        break;

                    case 0x11: // last multicast status
                        data.data[0] = multicastCmd;
                        data.data[1] = multicastSequence;
                        data.data[2] = multicastResult;
                        data.len = 3;
                        break;

                    case 0x12:  // get white calibration
                        {
                            uint8_t i, j;
                            uint16_t *ptr;

                            ptr = (uint16_t*)data.data;
                            for (i = 0; i < 6; i++)
                            {
                                for (j = 0; j < 3; j++)
                                {
                                    *ptr++ = whitetab[i][j];
                                }
                            }
                            data.len = 6 * 3 * sizeof(uint16_t);
                        }
                        break;

                    case 0x13: // set white calibration
                        if (len < 6 * 3 * sizeof(uint16_t))
                        {
                            data.id |= ID_ERR_PARAM;
                        }
                        else
                        {
                            uint8_t i, j;
                            uint16_t *ptr;

                            ptr = (uint16_t*)data.data;
                            for (i = 0; i < 6; i++)
                            {
                                for (j = 0; j < 3; j++)
                                {
                                    whitetab[i][j] = *ptr++;
                                }
                            }
                        }
                        break;

                    case 0x14:  // save white calibration
                        saveEeprom((uint8_t*)eep_whitetab, (uint8_t*)whitetab, sizeof(whitetab));
                        break;

                    case 0x15:  // get gamma correction
                        memcpy(data.data, gammatab, sizeof(gammatab));
                        data.len = sizeof(gammatab);
                        break;

                    case 0x16:  // set gamma correction
                        if (len < sizeof(gammatab))
                        {
                            data.id |= ID_ERR_PARAM;
                        }
                        else
                        {
                            uint8_t i;
                            uint16_t* ptr;

                            ptr = (uint16_t*)data.data;
                            for (i = 0; i < 33; i++)
                            {
                                if ((i == 0 && ptr[0] != 0) ||
                                    (i == 32 && ptr[32] != 0x6000) ||
                                    (i > 0 && ptr[i - 1] > ptr[i]))
                                {
                                    break;
                                }
                            }
                            if (i < 33)
                            {
                                data.id |= ID_ERR_PARAM;
                            }
                            else
                            {
                                memcpy(gammatab, data.data, sizeof(gammatab));
                            }
                        }
                        break;

                    case 0x17:  // save gamma correction
                        saveEeprom((uint8_t*)eep_gammatab, (uint8_t*)gammatab, sizeof(gammatab));
                        break;

                    case 0x18:  // versions string
                        memcpy_P(data.data, version, sizeof(version));
                        data.len = sizeof(version);
                        break;

                    case 0x19:  // get logical configuration
                        memcpy(data.data, config, sizeof(config));
                        data.len = sizeof(config);
                        break;

                    case 0x1a:  // set logical configuration
                        if (len < sizeof(config))
                        {
                            data.id |= ID_ERR_PARAM;
                        }
                        else
                        {
                            memcpy(config, data.data, sizeof(config));
                        }
                        break;

                    case 0x1b:  // save logical configuration
                        saveEeprom((uint8_t*)eep_config, (uint8_t*)config, sizeof(config));
                        break;

                    case 0x30:  // Performance test: calculate 1000x a crc
                        {
                            uint16_t start, stop;
                            uint8_t sreg;
                            uint8_t i, j;
                            uint8_t crc = 0;
                            uint8_t* ptr;

                            for (i = 0; i < 200; i++)
                            {
                                data.data[i] = i;
                            }

                            sreg = SREG;
                            cli();
                            start = TCNT1;
                            SREG = sreg;

                            for (i = 0; i < 5; i++)
                            {
                                ptr = data.data;
                                for (j = 0; j < 200; j++)
                                {
                                    crc = _crc_ibutton_update(crc, *ptr++);
                                }
                            }

                            sreg = SREG;
                            cli();
                            stop = TCNT1;
                            SREG = sreg;

                            if (stop <= start)
                            {
                                stop += 0x6000;
                            }
                            stop -= start;

                            data.data[0] = stop % 256;
                            data.data[1] = stop / 256;
                            data.len = 2;
                        }
                        break;

                    case 0x31: // Performance test: calculate 5x new rgb values
                        {
                            uint16_t start, stop;
                            uint8_t i;
                            uint8_t sreg;

                            for (i = 0; i < 6; i++)
                            {
                                data.data[4 * i] = (id << 3) + i;
                                data.data[4 * i + 1] = i * 30 + 10; 
                                data.data[4 * i + 2] = i * 30 + 20; 
                                data.data[4 * i + 3] = i * 30 + 30; 
                            }
                            data.len = 6 * 4;

                            sreg = SREG;
                            cli();
                            start = TCNT1;
                            SREG = sreg;

                            for (i = 0; i < 5; i++)
                            {
                                ptr = getRgbValues(0);
                                
                                sreg = SREG;
				                cli();
				                ptr_next = ptr;
				                SREG = sreg;
                                ptr = NULL;
                            }
                             
                            sreg = SREG;
                            cli();
                            stop = TCNT1;
                            SREG = sreg;

                            if (stop <= start)
                            {
                                stop += 0x6000;
                            }
                            stop -= start;
                            data.data[0] = stop % 256;
                            data.data[1] = stop / 256;
                            data.len = 2;
                        }
                        break;
                }
            }
            data.cmd |= MSG_ANSWER;
            putMessage();
        }
    }
}


